from typing import Dict
from workspace import env, message, git
import os
from colorama import Fore, Back, Style

toplevel_cmake = """
cmake_minimum_required(VERSION 3.16.3)
project(wks)

# Libs are in lib/ and binaries in bin/
include(GNUInstallDirs)
set(CMAKE_RUNTIME_OUTPUT_DIRECTORY ${CMAKE_CURRENT_BINARY_DIR}/${CMAKE_INSTALL_BINDIR})
set(CMAKE_LIBRARY_OUTPUT_DIRECTORY ${CMAKE_CURRENT_BINARY_DIR}/${CMAKE_INSTALL_LIBDIR})
set(CMAKE_ARCHIVE_OUTPUT_DIRECTORY ${CMAKE_CURRENT_BINARY_DIR}/${CMAKE_INSTALL_LIBDIR})

add_subdirectory(src)
"""

cmake_headers = """
cmake_minimum_required(VERSION 3.16.3)

{PREFIX_PATHS}

# Colored compiler output
add_compile_options(
  $<$<CXX_COMPILER_ID:GNU>:-fdiagnostics-color=always>
  $<$<CXX_COMPILER_ID:Clang>:-fcolor-diagnostics>
)

# Easy build type selection
set_property(CACHE CMAKE_BUILD_TYPE PROPERTY STRINGS Release Debug RelWithDebInfo)

# Generate compile_commands.json to make it easier to work with clang based tools
set(CMAKE_EXPORT_COMPILE_COMMANDS ON)

# Create profile for profiling with gprof
if(CMAKE_CXX_COMPILER_ID STREQUAL "GNU")
  set(CMAKE_C_FLAGS_PROFILE
    "${CMAKE_C_FLAGS_RELWITHDEBINFO} -p" CACHE STRING "")
  set(CMAKE_CXX_FLAGS_PROFILE
    "${CMAKE_CXX_FLAGS_RELWITHDEBINFO} -p" CACHE STRING "")
  set(CMAKE_EXE_LINKER_FLAGS_PROFILE
    "${CMAKE_EXE_LINKER_FLAGS_RELWITHDEBINFO} -p" CACHE STRING "")
  set(CMAKE_SHARED_LINKER_FLAGS_PROFILE
    "${CMAKE_SHARED_LINKER_FLAGS_RELWITHDEBINFO} -p" CACHE STRING "")
  set(CMAKE_STATIC_LINKER_FLAGS_PROFILE
    "${CMAKE_STATIC_LINKER_FLAGS_RELWITHDEBINFO}" CACHE STRING "")
  set(CMAKE_MODULE_LINKER_FLAGS_PROFILE
    "${CMAKE_MODULE_LINKER_FLAGS_RELWITHDEBINFO} -p" CACHE STRING "")
  mark_as_advanced(FORCE CMAKE_C_FLAGS_PROFILE CMAKE_CXX_FLAGS_PROFILE CMAKE_EXE_LINKER_FLAGS_PROFILE CMAKE_SHARED_LINKER_FLAGS_PROFILE CMAKE_STATIC_LINKER_FLAGS_PROFILE CMAKE_MODULE_LINKER_FLAGS_PROFILE)
  ## set postfix for binaries
  set(CMAKE_PROFILE_POSTFIX _profile)

  ## Add Profile to selectable build types
  set_property(CACHE CMAKE_BUILD_TYPE APPEND PROPERTY STRINGS Profile)
endif()


"""


def generate():
    message.bright("* Generating CMake")
    print("")

    cmake = cmake_headers

    def add_cmake(directory):
        nonlocal cmake

        cmakes = [""]
        config = env.get_config(directory)
        if config and "cmakes" in config:
            cmakes = config["cmakes"]

        for cmake_name in cmakes:
            cmake_directory = os.path.realpath(directory)

            dname = os.path.basename(os.path.dirname(directory))
            name = os.path.basename(directory)
            project = "%s/%s" % (dname, name)

            if cmake_name:
                project += "/" + cmake_name

            if os.path.isfile(cmake_directory + "/" + cmake_name + "/CMakeLists.txt"):
                cmake += "add_subdirectory(%s)\n" % (project)
                print("- %s [%s]" % (project, Style.DIM + cmake_directory + Style.RESET_ALL))

    added = set()
    prefix_path = []

    def add_project(directory):
        nonlocal added

        if directory not in added:
            added.add(directory)
            config = env.get_config(directory)
            if config and "cmake_prefixes" in config:
                for entry in config["cmake_prefixes"]:
                    prefix_path.append(directory + "/" + entry)

            dependencies = []

            if config and "deps" in config:
                dependencies += config["deps"]
            if config and "optional" in config:
                dependencies += config["optional"]
            
            for dep in dependencies:
                dep_directory = git.parse_repository_name(dep)["directory"]
                add_project(dep_directory)

            add_cmake(directory)

    for directory in git.get_directories():
        add_project(directory)

    prefix_paths = ""
    for entry in prefix_path:
        prefix_paths += 'list(APPEND CMAKE_PREFIX_PATH "' + entry + '")\n'
    cmake = cmake.replace("{PREFIX_PATHS}", prefix_paths)

    f = open(env.sources_directory + "/CMakeLists.txt", "w")
    f.write(cmake)
    f.close()

    if not os.path.exists("CMakeLists.txt"):
        message.bright("* No top-level CMakeLists.txt found, generating one...")
        f = open("CMakeLists.txt", "w")
        f.write(toplevel_cmake)
        f.close()

    message.bright("* Wrote CMakeLists.txt")
    print("")



import os
import yaml

sources_directory = os.getcwd() + "/src"


def get_config(directory):
    deps_filename = directory + "/wks.yml"
    if os.path.exists(deps_filename):
        f = open(deps_filename, "r")
        deps = yaml.load(f, Loader=yaml.Loader)

        return deps

    return None



import os
from colorama import Fore, Back, Style


def error(message):
    print(Style.BRIGHT + Fore.RED + message + Style.RESET_ALL)


def bright(message):
    print("")
    print(Style.BRIGHT + message + Style.RESET_ALL)


def run_or_fail(cmd):
    if os.system(cmd) != 0:
        die("Error while running " + cmd)


def emphasis(text):
    return Fore.BLUE + text + Fore.RESET


def warn(text):
    return Fore.YELLOW + text + Fore.RESET


def error(text):
    return Fore.RED + text + Fore.RESET


def success(text):
    return Fore.GREEN + text + Fore.RESET


def die(message):
    print(error(message))
    exit()



import sys

from workspace import (
    command_build,
    command_clean,
    command_cmake,
    command_help,
    command_install,
    command_pull,
    command_status,
    command_graph,
    command_cmd,
)

# Adding all imported commands to the commands list
commands = []
entries = locals().copy()
for name in entries:
    if name.startswith("command_"):
        commands.append(entries[name])



import os
import subprocess
from colorama import Fore, Back, Style
from workspace import env, git, message, cmake

name = "status"
help = """Shows status for projects"""
usage = ""
min_args = 0


def run(args):
    config = env.get_config(os.getcwd())

    group_not_sync = []
    group_untracked_file = []
    group_modified_file = []

    for directory in git.get_directories():
        parts = directory.split("/")
        vendor = parts[-2]
        if config is None or vendor in config["status"]:
            branch = git.branch(directory)
            message_header = "* In " + directory + " [" + message.emphasis(branch) + "]"

            status = git.status(directory).strip()
            if not status:
                message.bright(message_header)
                print(message.success("- Everything is ok"))
            else:
                lines = status.split("\n")
                has_modified = False
                for line in lines:
                    parts = line.strip().split(" ")
                    if parts[0] != "??":
                        has_modified = True
                        break

                if has_modified:
                    group_modified_file.append(message_header)
                else:
                    group_untracked_file.append(message_header)

            if git.is_ahead(directory):
                group_not_sync.append(message_header)

    for m in group_not_sync:
        message.bright(m)
        print(message.error("- Is not in sync with remote"))

    for m in group_modified_file:
        message.bright(m)
        print(message.warn("- Has modified files"))

    for m in group_untracked_file:
        message.bright(m)
        print(message.error("- Has untracked files"))

    print("")



import os
from workspace import env, git, message, cmake

name = "pull"
help = """Pull all the repositories"""
usage = ""
min_args = 0


def run(args):
    vendor_filter = None if (len(args) == 0) else args[0]
    git.global_command("git pull", vendor_filter)
    git.scan_all_dependencies()
    cmake.generate()



import subprocess
import os
from workspace import env, git, message, cmake

name = "clean"
help = """Cleans the build"""
usage = ""
min_args = 0


def run(args):
    message.bright("* Cleaning")

    if subprocess.getoutput("which ninja") != "":
        cmd = "cd build; ninja clean"
        os.system(cmd)
    else:
        print(message.warn("Ninja was not found, you can consider: apt-get install ninja-build"))
        if input("Continue with Makefiles? (y/n) ") == "y":
            cmd = "cd build; make clean"
            os.system(cmd)

    message.bright(message.success("\nBuild was successful\n"))



import subprocess
import os
from workspace import env, message, env
import threading


def parse_repository_name(name):
    parts = name.split("/")

    if len(parts) != 2:
        message.die("Bad repository name: " + name)

    repository = {"vendor": parts[0], "name": parts[1], "branch": None, "tag": None}

    markers = {"#": "branch", "@": "tag"}

    for marker in markers:
        if marker in repository["name"]:
            name, value = repository["name"].split(marker, 1)
            repository["name"] = name
            repository[markers[marker]] = value

    repository["directory"] = env.sources_directory + "/" + repository["vendor"] + "/" + repository["name"]
    repository["git"] = "git@github.com:%s/%s.git" % (repository["vendor"], repository["name"])
    repository["full_name"] = "%s/%s" % (repository["vendor"], repository["name"])

    return repository


def install(repository_name, source):
    repository = parse_repository_name(repository_name)

    if not os.path.exists(repository["directory"]):
        message.bright("* Installing " + repository_name + "...")
        vendor_dir = os.path.dirname(repository["directory"])

        if not os.path.exists(vendor_dir):
            os.makedirs(vendor_dir)

        branchName = ""
        if repository["branch"]:
            branchName = "--branch " + repository["branch"]
        if repository["tag"]:
            branchName = "--branch " + repository["tag"]

        cmd = "git clone %s %s %s" % (branchName, repository["git"], repository["directory"])

        message.run_or_fail(cmd)

        config = env.get_config(repository["directory"])
        if config and "install" in config:
            for command in config["install"]:
                cmd = "cd %s; %s" % (repository["directory"], command)
                message.run_or_fail(cmd)

        return True
    else:
        if repository["branch"]:
            current_branch = branch(repository["directory"])
            if current_branch != repository["branch"]:
                print(
                    message.warn(
                        "WARNING: %s wants branch %s for %s, but it is currently %s"
                        % (source, repository["branch"], repository["full_name"], current_branch)
                    )
                )
        elif repository["tag"]:
            current_tag = tag(repository["directory"])
            if current_tag != repository["tag"]:
                print(
                    message.warn(
                        "WARNING: %s wants tag %s for %s, but it is currently %s"
                        % (source, repository["tag"], repository["full_name"], current_tag)
                    )
                )

        return False


def get_directories(directory=None):
    directories = []
    if directory is None:
        directory = env.sources_directory

    if not os.path.isdir(directory):
        message.die("Directory %s does not exists" % directory)

    subdirectories = [entry.name for entry in os.scandir(directory)]
    subdirectories.sort()

    for entry in subdirectories:
        full_name = directory + "/" + entry

        if os.path.isdir(full_name):
            if os.path.isdir(full_name + "/.git"):
                directories.append(full_name)
            else:
                directories += get_directories(full_name)

    return directories


def scan_dependencies(directory):
    changed = False
    config = env.get_config(directory)

    if config and "deps" in config:
        for entry in config["deps"]:
            if install(entry, os.path.basename(directory)):
                changed = True

    return changed


def scan_all_dependencies():
    message.bright("* Scanning all dependencies")
    changed = True

    while changed:
        changed = False
        for directory in get_directories():
            if scan_dependencies(directory):
                changed = True


def global_command(command, vendor_filter=None):
    message.bright("* Running global command: %s" % command)

    use_threads = os.getenv("WKS_NO_THREADS") is None
    directories = get_directories()

    threads: list[threading.Thread] = [None] * len(directories)
    processes: list[subprocess.CompletedProcess] = threads.copy()
    has_error: bool = False

    def thread_func(index: int, dir: str, cmd: str):
        processes[index] = subprocess.run(cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=dir)

    def show_output(process):
        if process.returncode != 0:
            print(message.error(process.stdout.decode()))
        else:
            print(process.stdout.decode())

    for index, directory in enumerate(directories):
        if vendor_filter is not None:
            parts = directory.split("/")
            vendor = parts[-2]
            if vendor.lower() != vendor_filter.lower():
                continue

        if use_threads:
            threads[index] = threading.Thread(None, thread_func, args=(index, directory, command))
            threads[index].start()
        else:
            message.bright("> %s" % directory)
            thread_func(index, directory, command)
            show_output(processes[index])
            if processes[index].returncode != 0:
                has_error = True

    if use_threads:
        for index, directory in enumerate(directories):
            threads[index].join()
            message.bright("- In %s ..." % os.path.realpath(directory))
            show_output(processes[index])
            if processes[index].returncode != 0:
                has_error = True

    print("")
    if has_error:
        print(message.error("Some commands failed"))


def status(directory):
    output = subprocess.getoutput("cd %s; git status --porcelain=v1" % directory)
    return output


def branch(directory):
    output = subprocess.getoutput("cd %s; git rev-parse --abbrev-ref HEAD" % directory)
    return output


def tag(directory):
    output = subprocess.getoutput("cd %s; git describe" % directory)
    return output


def is_ahead(directory):
    ret, output = subprocess.getstatusoutput("cd %s; git rev-list @{u}.." % directory)
    return ret == 0 and output.strip() != ""
